#!/bin/sh
#
# Copyright (C) 2010-2012 Denis Roio <jaromil@dyne.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# here below only auxiliary functions

get_conf() {
    grep "^$1" /conf/config |  perl -F\' -anale 'print $F[1]'
}

get_ip() {
    if [ "$HDSYNC_NETWORK" = "DYNAMIC" ]; then
	echo "`date +%T` listening for DHCP assigned IP on the network"
	config_tool -c LAN_TYPE='d'
	IP=`get_conf IP2`
    elif [ "$HDSYNC_NETWORK" = "STATIC" ]; then
	echo "`date +%T` setting a static network address"
	IP="192.168.0.$HDSYNC_CHANNEL"
	config_tool -c LAN_TYPE='s'
	config_tool -c IP2=$IP
	config_tool -c NETMASK2=255.255.255.0
	# loopback route
	route add -net 127.0.0.0 netmask 255.0.0.0 dev lo
	# local network address
	ifconfig eth0 $IP netmask 255.255.255.0
	# from the route(8) manual page: "This is an obscure one
	# documented so people know how to do it. This sets all of the
	# class D (multicast) IP routes to go via "eth0". This is the
	# correct normal configuration line with a multicasting
	# kernel."
	route add -net 224.0.0.0 netmask 240.0.0.0 dev eth0
	# and this is also necessary, default gateway is own interface
	# (usage is compatible with some oldish BSD notation)
	route add default gw $IP

	echo "`date +%T` network interface configured with address $IP ..."
    else
	IP=`get_conf IP2`
	echo "`date +%T` network interface manualy configured with address $IP ..."
    fi
    export IP
}


